@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -75,15 +75,20 @@ int main(int argc, char *argv[])
 	int force = 0;
 	int test = 0;
 	int assume_clean = 0;
-	int autof = 0; /* -2 means create device based on name:
-			*    if it ends mdN, then non-partitioned array N
-			*    if it ends dN, then partitions array N
-			* -1 means create non-partitioned, choose N
-			*  1 or more to create partitioned
-			* If -1 or 1 and name is a 'standard' name, then
-			* insist on a match of type and number.
-			*/
+	/* autof indicates whether and how to create device node.
+	 * bottom 3 bits are style.  Rest (when shifted) are number of parts
+	 * 0  - unset
+	 * 1  - don't create (no)
+	 * 2  - if is_standard, then create (yes)
+	 * 3  - create as 'md' - reject is_standard mdp (md)
+	 * 4  - create as 'mdp' - reject is_standard md (mdp)
+	 * 5  - default to md if not is_standard (md in config file)
+	 * 6  - default to mdp if not is_standard (part, or mdp in config file)
+	 */
+	int autof = 0;
 
+	char *homehost = NULL;
+	char sys_hostname[256];
 	char *mailaddr = NULL;
 	char *program = NULL;
 	int delay = 0;
@@ -95,6 +100,7 @@ int main(int argc, char *argv[])
 	int re_add = 0;
 	char *shortopt = short_options;
 	int dosyslog = 0;
+	int auto_update_home = 0;
 
 	int copies;
 	int print_help = 0;
@@ -150,6 +156,10 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case HomeHost:
+			homehost = optarg;
+			continue;
+
 		case ':':
 		case '?':
 			fputs(Usage, stderr);
@@ -168,7 +178,7 @@ int main(int argc, char *argv[])
 		case 'a':
 		case 'r':
 		case 'f':
-		case 6: /* re-add */
+		case ReAdd: /* re-add */
 			if (!mode) {
 				newmode = MANAGE;
 				shortopt = short_bitmap_auto_options;
@@ -228,6 +238,7 @@ int main(int argc, char *argv[])
 					dv->disposition = devmode;
 					dv->writemostly = writemostly;
 					dv->re_add = re_add;
+					dv->used = 0;
 					dv->next = NULL;
 					*devlistend = dv;
 					devlistend = &dv->next;
@@ -305,6 +316,9 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case O(ASSEMBLE,AutoHomeHost):
+			auto_update_home = 1;
+			continue;
 		case O(CREATE,'e'):
 		case O(ASSEMBLE,'e'):
 		case O(MISC,'e'): /* set metadata (superblock) information */
@@ -403,16 +417,18 @@ int main(int argc, char *argv[])
 				break;
 
 			case 10:
-				/* 'f' or 'n' followed by a number <= raid_disks */
-				if ((optarg[0] !=  'n' && optarg[0] != 'f') ||
+				/* 'f', 'o' or 'n' followed by a number <= raid_disks */
+				if ((optarg[0] !=  'n' && optarg[0] != 'f' && optarg[0] != 'o') ||
 				    (copies = strtoul(optarg+1, &cp, 10)) < 1 ||
 				    copies > 200 ||
 				    *cp) {
-					fprintf(stderr, Name ": layout for raid10 must be 'nNN' or 'fNN' where NN is a number, not %s\n", optarg);
+					fprintf(stderr, Name ": layout for raid10 must be 'nNN', 'oNN' or 'fNN' where NN is a number, not %s\n", optarg);
 					exit(2);
 				}
 				if (optarg[0] == 'n')
 					layout = 256 + copies;
+				else if (optarg[0] == 'o')
+					layout = 0x10000 + (copies<<8) + 1;
 				else
 					layout = 1 + (copies<<8);
 				break;
@@ -436,8 +452,8 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
-		case O(CREATE,3):
-		case O(BUILD,3): /* assume clean */
+		case O(CREATE,AssumeClean):
+		case O(BUILD,AssumeClean): /* assume clean */
 			assume_clean = 1;
 			continue;
 
@@ -480,39 +496,7 @@ int main(int argc, char *argv[])
 		case O(CREATE,'a'):
 		case O(BUILD,'a'):
 		case O(ASSEMBLE,'a'): /* auto-creation of device node */
-			if (optarg == NULL)
-				autof = -2;
-			else if (strcasecmp(optarg,"no")==0)
-				autof = 0;
-			else if (strcasecmp(optarg,"yes")==0)
-				autof = -2;
-			else if (strcasecmp(optarg,"md")==0)
-				autof = -1;
-			else {
-				/* There might be digits, and maybe a hypen, at the end */
-				char *e = optarg + strlen(optarg);
-				int num = 4;
-				int len;
-				while (e > optarg && isdigit(e[-1]))
-					e--;
-				if (*e) {
-					num = atoi(e);
-					if (num <= 0) num = 1;
-				}
-				if (e > optarg && e[-1] == '-')
-					e--;
-				len = e - optarg;
-				if ((len == 3 && strncasecmp(optarg,"mdp",3)==0) ||
-				    (len == 1 && strncasecmp(optarg,"p",1)==0) ||
-				    (len >= 4 && strncasecmp(optarg,"part",4)==0))
-					autof = num;
-				else {
-					fprintf(stderr, Name ": --auto flag arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
-						"        optionally followed by a number.\n",
-						optarg);
-					exit(2);
-				}
-			}
+			autof = parse_auto(optarg, "--auto flag", 0);
 			continue;
 
 		case O(BUILD,'f'): /* force honouring '-n 1' */
@@ -587,6 +571,10 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "uuid")==0)
 				continue;
+			if (strcmp(update, "name")==0)
+				continue;
+			if (strcmp(update, "homehost")==0)
+				continue;
 			if (strcmp(update, "byteorder")==0) {
 				if (ss) {
 					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
@@ -604,6 +592,10 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'uuid', 'resync' or 'summaries' supported\n",update);
 			exit(2);
 
+		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
+			runstop = -1; /* --stop isn't allowed for --assemble, so we overload slightly */
+			continue;
+
 		case O(ASSEMBLE,'c'): /* config file */
 		case O(MISC, 'c'):
 		case O(MONITOR,'c'):
@@ -613,6 +605,7 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			configfile = optarg;
+			set_conffile(configfile);
 			/* FIXME possibly check that config file exists.  Even parse it */
 			continue;
 		case O(ASSEMBLE,'s'): /* scan */
@@ -682,7 +675,7 @@ int main(int argc, char *argv[])
 			devmode = 'a';
 			re_add = 0;
 			continue;
-		case O(MANAGE,6):
+		case O(MANAGE,ReAdd):
 			devmode = 'a';
 			re_add = 1;
 			continue;
@@ -746,7 +739,7 @@ int main(int argc, char *argv[])
 			test = 1;
 			continue;
 
-		case O(MISC, 22):
+		case O(MISC, Sparc22):
 			if (devmode != 'E') {
 				fprintf(stderr, Name ": --sparc2.2 only allowed with --examine\n");
 				exit(2);
@@ -771,8 +764,8 @@ int main(int argc, char *argv[])
 			ident.bitmap_fd = bitmap_fd; /* for Assemble */
 			continue;
 
-		case O(ASSEMBLE, 7):
-		case O(GROW, 7):
+		case O(ASSEMBLE, BackupFile):
+		case O(GROW, BackupFile):
 			/* Specify a file into which grow might place a backup,
 			 * or from which assemble might recover a backup
 			 */
@@ -796,9 +789,9 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n");
 			exit(2);
 
-		case O(GROW,4):
-		case O(BUILD,4):
-		case O(CREATE,4): /* bitmap chunksize */
+		case O(GROW,BitmapChunk):
+		case O(BUILD,BitmapChunk):
+		case O(CREATE,BitmapChunk): /* bitmap chunksize */
 			bitmap_chunk = strtol(optarg, &c, 10);
 			if (!optarg[0] || *c || bitmap_chunk < 0 ||
 					bitmap_chunk & (bitmap_chunk - 1)) {
@@ -810,8 +803,8 @@ int main(int argc, char *argv[])
 			bitmap_chunk = bitmap_chunk ? bitmap_chunk * 1024 : 512;
 			continue;
 
-		case O(BUILD, 5):
-		case O(CREATE, 5): /* write-behind mode */
+		case O(BUILD, WriteBehind):
+		case O(CREATE, WriteBehind): /* write-behind mode */
 			write_behind = DEFAULT_MAX_WRITE_BEHIND;
 			if (optarg) {
 				write_behind = strtol(optarg, &c, 10);
@@ -885,6 +878,8 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");	
 			exit(2);
 		}
+		if (mode == MANAGE || mode == GROW)
+			autof=1; /* Don't create */
 		mdfd = open_mddev(devlist->devname, autof);
 		if (mdfd < 0)
 			exit(1);
@@ -916,6 +911,15 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (homehost == NULL)
+		homehost = conf_get_homehost();
+	if (homehost && strcmp(homehost, "<system>")==0) {
+		if (gethostname(sys_hostname, sizeof(sys_hostname)) == 0) {
+			sys_hostname[sizeof(sys_hostname)-1] = 0;
+			homehost = sys_hostname;
+		}
+	}
+
 	rv = 0;
 	switch(mode) {
 	case MANAGE:
@@ -928,13 +932,13 @@ int main(int argc, char *argv[])
 		if (!rv && readonly < 0)
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && runstop)
-			rv = Manage_runstop(devlist->devname, mdfd, runstop, 0);
+			rv = Manage_runstop(devlist->devname, mdfd, runstop, quiet);
 		break;
 	case ASSEMBLE:
 		if (devs_found == 1 && ident.uuid_set == 0 &&
 		    ident.super_minor == UnSet && ident.name[0] == 0 && !scan ) {
 			/* Only a device has been given, so get details from config file */
-			mddev_ident_t array_ident = conf_get_ident(configfile, devlist->devname);
+			mddev_ident_t array_ident = conf_get_ident(devlist->devname);
 			if (array_ident == NULL) {
 				fprintf(stderr, Name ": %s not identified in config file.\n",
 					devlist->devname);
@@ -945,16 +949,16 @@ int main(int argc, char *argv[])
 				if (mdfd < 0)
 					rv |= 1;
 				else {
-					rv |= Assemble(ss, devlist->devname, mdfd, array_ident, configfile,
+					rv |= Assemble(ss, devlist->devname, mdfd, array_ident,
 						       NULL, backup_file,
-						       readonly, runstop, update, verbose-quiet, force);
+						       readonly, runstop, update, homehost, verbose-quiet, force);
 					close(mdfd);
 				}
 			}
 		} else if (!scan)
-			rv = Assemble(ss, devlist->devname, mdfd, &ident, configfile,
+			rv = Assemble(ss, devlist->devname, mdfd, &ident,
 				      devlist->next, backup_file,
-				      readonly, runstop, update, verbose-quiet, force);
+				      readonly, runstop, update, homehost, verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
@@ -965,7 +969,7 @@ int main(int argc, char *argv[])
 				exit(1);
 			}
 			for (dv = devlist ; dv ; dv=dv->next) {
-				mddev_ident_t array_ident = conf_get_ident(configfile, dv->devname);
+				mddev_ident_t array_ident = conf_get_ident(dv->devname);
 				if (array_ident == NULL) {
 					fprintf(stderr, Name ": %s not identified in config file.\n",
 						dv->devname);
@@ -978,47 +982,104 @@ int main(int argc, char *argv[])
 					rv |= 1;
 					continue;
 				}
-				rv |= Assemble(ss, dv->devname, mdfd, array_ident, configfile,
+				rv |= Assemble(ss, dv->devname, mdfd, array_ident,
 					       NULL, backup_file,
-					       readonly, runstop, update, verbose-quiet, force);
+					       readonly, runstop, update, homehost, verbose-quiet, force);
 				close(mdfd);
 			}
 		} else {
-			mddev_ident_t array_list =  conf_get_ident(configfile, NULL);
-			if (!array_list) {
+			mddev_ident_t array_list =  conf_get_ident(NULL);
+			mddev_dev_t devlist = conf_get_devs();
+			int cnt = 0;
+			if (devlist == NULL) {
+				fprintf(stderr, Name ": No devices listed in conf file were found.\n");
+				exit(1);
+			}
+			if (update) {
+				fprintf(stderr, Name ": --update not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			if (backup_file) {
+				fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			for (; array_list; array_list = array_list->next) {
+				mdu_array_info_t array;
+				mdfd = open_mddev(array_list->devname,
+						  array_list->autof ? array_list->autof : autof);
+				if (mdfd < 0) {
+					rv |= 1;
+					continue;
+				}
+				if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
+					/* already assembled, skip */
+					;
+				else {
+					rv |= Assemble(ss, array_list->devname, mdfd,
+						       array_list,
+						       devlist, NULL,
+						       readonly, runstop, NULL, homehost, verbose-quiet, force);
+					if (rv == 0) cnt++;
+				}
+				close(mdfd);
+			}
+			if (homehost) {
+				/* Maybe we can auto-assemble something.
+				 * Repeatedly call Assemble in auto-assmble mode
+				 * until it fails
+				 */
+				int rv2;
+				int acnt;
+				ident.autof = autof;
+				do {
+					acnt = 0;
+					do {
+						rv2 = Assemble(ss, NULL, -1,
+							       &ident,
+							       devlist, NULL,
+							       readonly, runstop, NULL, homehost, verbose-quiet, force);
+						if (rv2==0) {
+							cnt++;
+							acnt++;
+						}
+						if (rv2 == 1)
+							/* found something so even though assembly failed  we
+							 * want to avoid auto-updates
+							 */
+							auto_update_home = 0;
+					} while (rv2!=2);
+					/* Incase there are stacked devices, we need to go around again */
+					devlist = conf_get_devs();
+				} while (acnt);
+				if (cnt == 0 && auto_update_home && homehost) {
+					/* Nothing found, maybe we need to bootstrap homehost info */
+					do {
+						acnt = 0;
+						do {
+							rv2 = Assemble(ss, NULL, -1,
+								       &ident,
+								       devlist, NULL,
+								       readonly, runstop, "homehost", homehost, verbose-quiet, force);
+							if (rv2==0) {
+								cnt++;
+								acnt++;
+							}
+						} while (rv2!=2);
+						/* Incase there are stacked devices, we need to go around again */
+						devlist = conf_get_devs();
+					} while (acnt);
+				}
+				if (cnt == 0 && rv == 0) {
+					fprintf(stderr, Name ": No arrays found in config file or automatically\n");
+					rv = 1;
+				}
+			} else if (cnt == 0 && rv == 0) {
 				fprintf(stderr, Name ": No arrays found in config file\n");
 				rv = 1;
-			} else
-				if (update) {
-					fprintf(stderr, Name ": --update not meaningful with a --scan assembly.\n");
-					exit(1);
-				}
-				if (backup_file) {
-					fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
-					exit(1);
-				}
-				for (; array_list; array_list = array_list->next) {
-					mdu_array_info_t array;
-					mdfd = open_mddev(array_list->devname, 
-							  array_list->autof ? array_list->autof : autof);
-					if (mdfd < 0) {
-						rv |= 1;
-						continue;
-					}
-					if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
-						/* already assembled, skip */
-						;
-					else
-						rv |= Assemble(ss, array_list->devname, mdfd,
-							       array_list, configfile,
-							       NULL, NULL,
-							       readonly, runstop, NULL, verbose-quiet, force);
-					close(mdfd);
-				}
+			}
 		}
 		break;
 	case BUILD:
-		if (bitmap_chunk == UnSet) bitmap_chunk = DEFAULT_BITMAP_CHUNK;
 		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 		if (write_behind && !bitmap_file) {
 			fprintf(stderr, Name ": write-behind mode requires a bitmap.\n");
@@ -1046,7 +1107,7 @@ int main(int argc, char *argv[])
 		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
-			    raiddisks, sparedisks, ident.name,
+			    raiddisks, sparedisks, ident.name, homehost,
 			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
 			    bitmap_file, bitmap_chunk, write_behind, delay);
 		break;
@@ -1058,12 +1119,12 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			if (devlist == NULL)
-				devlist = conf_get_devs(configfile);
+				devlist = conf_get_devs();
 			if (devlist == NULL) {
 				fprintf(stderr, Name ": No devices listed in %s\n", configfile?configfile:DefaultConfFile);
 				exit(1);
 			}
-			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss);
+			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss, homehost);
 		} else {
 			if (devlist == NULL) {
 				if (devmode=='D' && scan) {
@@ -1078,7 +1139,7 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1, test);
+						rv |= Detail(name, verbose>1?0:verbose+1, test, homehost);
 						put_md_name(name);
 					}
 				} else	if (devmode == 'S' && scan) {
@@ -1102,9 +1163,9 @@ int main(int argc, char *argv[])
 									e->dev);
 								continue;
 							}
-							mdfd = open_mddev(name, 0);
+							mdfd = open_mddev(name, 1);
 							if (mdfd >= 0) {
-								if (Manage_runstop(name, mdfd, -1, !last))
+								if (Manage_runstop(name, mdfd, -1, quiet?1:last?0:-1))
 									err = 1;
 								else
 									progress = 1;
@@ -1123,7 +1184,7 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief?1+verbose:0, test); continue;
+					rv |= Detail(dv->devname, brief?1+verbose:0, test, homehost); continue;
 				case 'K': /* Zero superblock */
 					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
@@ -1131,13 +1192,13 @@ int main(int argc, char *argv[])
 				case 'X':
 					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
 				}
-				mdfd = open_mddev(dv->devname, 0);
+				mdfd = open_mddev(dv->devname, 1);
 				if (mdfd>=0) {
 					switch(dv->disposition) {
 					case 'R':
-						rv |= Manage_runstop(dv->devname, mdfd, 1, 0); break;
+						rv |= Manage_runstop(dv->devname, mdfd, 1, quiet); break;
 					case 'S':
-						rv |= Manage_runstop(dv->devname, mdfd, -1, 0); break;
+						rv |= Manage_runstop(dv->devname, mdfd, -1, quiet); break;
 					case 'o':
 						rv |= Manage_ro(dv->devname, mdfd, 1); break;
 					case 'w':
@@ -1162,7 +1223,7 @@ int main(int argc, char *argv[])
 		}
 		rv= Monitor(devlist, mailaddr, program,
 			    delay?delay:60, daemonise, scan, oneshot,
-			    dosyslog, configfile, test, pidfile);
+			    dosyslog, test, pidfile);
 		break;
 
 	case GROW:
