@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -27,7 +27,7 @@
  *           Australia
  */
 
-#define	__USE_LARGEFILE64
+#define	_GNU_SOURCE
 #include	<unistd.h>
 #ifndef __dietlibc__
 extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
@@ -44,10 +44,11 @@ extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
 #include	<errno.h>
 #include	<string.h>
 #include	<syslog.h>
-#ifdef __dietlibc__NONO
-int strncmp(const char *s1, const char *s2, size_t n) __THROW __pure__;
-char *strncpy(char *dest, const char *src, size_t n) __THROW;
-#include    <strings.h>
+#ifdef __dietlibc__
+#include	<strings.h>
+/* dietlibc has deprecated random and srandom!! */
+#define random rand
+#define srandom srand
 #endif
 
 
@@ -71,12 +72,67 @@ char *strncpy(char *dest, const char *src, size_t n) __THROW;
 #include	"md_p.h"
 #include	"bitmap.h"
 
+#include <endian.h>
+/* Redhat don't like to #include <asm/byteorder.h>, and
+ * some time include <linux/byteorder/xxx_endian.h> isn't enough,
+ * and there is no standard conversion function so... */
+/* And dietlibc doesn't think byteswap is ok, so.. */
+/*  #include <byteswap.h> */
+#define bswap_16(x) (((x) & 0x00ffU) << 8 | \
+		     ((x) & 0xff00U) >> 8)
+#define bswap_32(x) (((x) & 0x000000ffU) << 24 | \
+		     ((x) & 0xff000000U) >> 24 | \
+		     ((x) & 0x0000ff00U) << 8  | \
+		     ((x) & 0x00ff0000U) >> 8)
+#define bswap_64(x) (((x) & 0x00000000000000ffULL) << 56 | \
+		     ((x) & 0xff00000000000000ULL) >> 56 | \
+		     ((x) & 0x000000000000ff00ULL) << 40 | \
+		     ((x) & 0x00ff000000000000ULL) >> 40 | \
+		     ((x) & 0x0000000000ff0000ULL) << 24 | \
+		     ((x) & 0x0000ff0000000000ULL) >> 24 | \
+		     ((x) & 0x00000000ff000000ULL) << 8 | \
+		     ((x) & 0x000000ff00000000ULL) >> 8)
+
+#if BYTE_ORDER == LITTLE_ENDIAN
+#define	__cpu_to_le16(_x) (_x)
+#define __cpu_to_le32(_x) (_x)
+#define __cpu_to_le64(_x) (_x)
+#define	__le16_to_cpu(_x) (_x)
+#define __le32_to_cpu(_x) (_x)
+#define __le64_to_cpu(_x) (_x)
+#elif BYTE_ORDER == BIG_ENDIAN
+#define	__cpu_to_le16(_x) bswap_16(_x)
+#define __cpu_to_le32(_x) bswap_32(_x)
+#define __cpu_to_le64(_x) bswap_64(_x)
+#define	__le16_to_cpu(_x) bswap_16(_x)
+#define __le32_to_cpu(_x) bswap_32(_x)
+#define __le64_to_cpu(_x) bswap_64(_x)
+#else
+#  error "unknown endianness."
+#endif
+
+
+
 /* general information that might be extracted from a superblock */
 struct mdinfo {
 	mdu_array_info_t	array;
 	mdu_disk_info_t		disk;
 	__u64			events;
 	int			uuid[4];
+	char			name[33];
+	unsigned long long	data_offset;
+	unsigned long long	component_size;
+	int			reshape_active;
+	unsigned long long	reshape_progress;
+	int			new_level, delta_disks, new_layout, new_chunk;
+};
+
+struct createinfo {
+	int	uid;
+	int	gid;
+	int	autof;
+	int	mode;
+	struct supertype *supertype;
 };
 
 #define Name "mdadm"
@@ -98,6 +154,21 @@ extern char Version[], Usage[], Help[], OptionHelp[],
 	Help_create[], Help_build[], Help_assemble[], Help_grow[],
 	Help_manage[], Help_misc[], Help_monitor[], Help_config[];
 
+/* for option that don't have short equivilents, we assign arbitrary
+ * small numbers.  '1' means an undecorated option, so we start at '2'.
+ */
+enum special_options {
+	AssumeClean = 2,
+	BitmapChunk,
+	WriteBehind,
+	ReAdd,
+	NoDegraded,
+	Sparc22,
+	BackupFile,
+	HomeHost,
+	AutoHomeHost,
+};
+
 /* structures read from config file */
 /* List of mddevice names and identifiers
  * Identifiers can be:
@@ -127,6 +198,7 @@ typedef struct mddev_ident_s {
 	struct supertype *st;
 	int	autof;		/* 1 for normal, 2 for partitioned */
 	char	*spare_group;
+	char	*bitmap_file;
 	int	bitmap_fd;
 
 	struct mddev_ident_s *next;
@@ -140,6 +212,7 @@ typedef struct mddev_dev_s {
 				 */
 	char writemostly;
 	char re_add;
+	char used;		/* set when used */
 	struct mddev_dev_s *next;
 } *mddev_dev_t;
 
@@ -164,6 +237,58 @@ extern struct mdstat_ent *mdstat_read(int hold, int start);
 extern void free_mdstat(struct mdstat_ent *ms);
 extern void mdstat_wait(int seconds);
 
+/* Data structure for holding info read from sysfs */
+struct sysdev {
+	char	name[20];
+	int	role;
+	int	major, minor;
+	unsigned long long offset, size;
+	int	state;
+	int	errors;
+	struct sysdev *next;
+};
+struct sysarray {
+	char	name[20];
+	struct sysdev *devs;
+	int	chunk;
+	unsigned long long component_size;
+	int	layout;
+	int	level;
+	int	spares;
+};
+/* various details can be requested */
+#define	GET_LEVEL	1
+#define	GET_LAYOUT	2
+#define	GET_COMPONENT	4
+#define	GET_CHUNK	8
+
+#define	GET_DEVS	1024 /* gets role, major, minor */
+#define	GET_OFFSET	2048
+#define	GET_SIZE	4096
+#define	GET_STATE	8192
+#define	GET_ERROR	16384
+
+/* If fd >= 0, get the array it is open on,
+ * else use devnum. >=0 -> major9. <0.....
+ */
+extern struct sysarray *sysfs_read(int fd, int devnum, unsigned long options);
+extern int sysfs_set_str(struct sysarray *sra, struct sysdev *dev,
+			 char *name, char *val);
+extern int sysfs_set_num(struct sysarray *sra, struct sysdev *dev,
+			 char *name, unsigned long long val);
+extern int sysfs_get_ll(struct sysarray *sra, struct sysdev *dev,
+			char *name, unsigned long long *val);
+
+
+extern int save_stripes(int *source, unsigned long long *offsets,
+			int raid_disks, int chunk_size, int level, int layout,
+			int nwrites, int *dest,
+			unsigned long long start, unsigned long long length);
+extern int restore_stripes(int *dest, unsigned long long *offsets,
+			   int raid_disks, int chunk_size, int level, int layout,
+			   int source, unsigned long long read_offset,
+			   unsigned long long start, unsigned long long length);
+
 #ifndef Sendmail
 #define Sendmail "/usr/lib/sendmail -t"
 #endif
@@ -174,19 +299,22 @@ extern char *map_num(mapping_t *map, int num);
 extern int map_name(mapping_t *map, char *name);
 extern mapping_t r5layout[], pers[], modes[], faultylayout[];
 
-extern char *map_dev(int major, int minor);
+extern char *map_dev(int major, int minor, int create);
 
 
 extern struct superswitch {
-	void (*examine_super)(void *sbv);
+	void (*examine_super)(void *sbv, char *homehost);
 	void (*brief_examine_super)(void *sbv);
-	void (*detail_super)(void *sbv);
+	void (*detail_super)(void *sbv, char *homehost);
 	void (*brief_detail_super)(void *sbv);
 	void (*uuid_from_super)(int uuid[4], void *sbv);
-	void (*getinfo_super)(struct mdinfo *info, mddev_ident_t ident, void *sbv);
-	int (*update_super)(struct mdinfo *info, void *sbv, char *update, char *devname, int verbose);
+	void (*getinfo_super)(struct mdinfo *info, void *sbv);
+	int (*match_home)(void *sbv, char *homehost);
+	int (*update_super)(struct mdinfo *info, void *sbv, char *update,
+			    char *devname, int verbose,
+			    int uuid_set, char *homehost);
 	__u64 (*event_super)(void *sbv);
-	int (*init_super)(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *name);
+	int (*init_super)(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *name, char *homehost);
 	void (*add_to_super)(void *sbv, mdu_disk_info_t *dinfo);
 	int (*store_super)(struct supertype *st, int fd, void *sbv);
 	int (*write_init_super)(struct supertype *st, void *sbv, mdu_disk_info_t *dinfo, char *devname);
@@ -215,17 +343,32 @@ extern struct supertype *guess_super(int fd);
 struct stat64;
 #endif
 
+#define HAVE_NFTW  we assume
+#define HAVE_FTW
+
 #ifdef UCLIBC
-  struct FTW {};
-# define FTW_PHYS 1
-#else
-# define  __USE_XOPEN_EXTENDED
-# include <ftw.h>
-# ifdef __dietlibc__
-#  define FTW_PHYS 1
+# include <features.h>
+# ifndef  __UCLIBC_HAS_FTW__
+#  undef HAVE_FTW
+#  undef HAVE_NFTW
 # endif
 #endif
 
+#ifdef __dietlibc__
+# undef HAVE_NFTW
+#endif
+
+#ifndef HAVE_NFTW
+# define FTW_PHYS 1
+# ifndef HAVE_FTW
+  struct FTW {};
+# endif
+#endif
+
+#ifdef HAVE_FTW
+# include <ftw.h>
+#endif
+
 extern int add_dev(const char *name, const struct stat *stb, int flag, struct FTW *s);
 
 
@@ -237,14 +380,18 @@ extern int Manage_subdevs(char *devname, int fd,
 			  mddev_dev_t devlist, int verbose);
 extern int Grow_Add_device(char *devname, int fd, char *newdev);
 extern int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int write_behind, int force);
+extern int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
+			long long size,
+			int level, int layout, int chunksize, int raid_disks);
+extern int Grow_restart(struct supertype *st, struct mdinfo *info,
+			int *fdlist, int cnt, char *backup_file);
 
 
 extern int Assemble(struct supertype *st, char *mddev, int mdfd,
 		    mddev_ident_t ident,
-		    char *conffile,
-		    mddev_dev_t devlist,
+		    mddev_dev_t devlist, char *backup_file,
 		    int readonly, int runstop,
-		    char *update,
+		    char *update, char *homehost,
 		    int verbose, int force);
 
 extern int Build(char *mddev, int mdfd, int chunk, int level, int layout,
@@ -255,21 +402,21 @@ extern int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 
 extern int Create(struct supertype *st, char *mddev, int mdfd,
 		  int chunk, int level, int layout, unsigned long long size, int raiddisks, int sparedisks,
-		  char *name,
+		  char *name, char *homehost,
 		  int subdevs, mddev_dev_t devlist,
 		  int runstop, int verbose, int force, int assume_clean,
 		  char *bitmap_file, int bitmap_chunk, int write_behind, int delay);
 
-extern int Detail(char *dev, int brief, int test);
+extern int Detail(char *dev, int brief, int test, char *homehost);
 extern int Query(char *dev);
 extern int Examine(mddev_dev_t devlist, int brief, int scan, int SparcAdjust,
-		   struct supertype *forcest);
+		   struct supertype *forcest, char *homehost);
 extern int Monitor(mddev_dev_t devlist,
 		   char *mailaddr, char *alert_cmd,
 		   int period, int daemonise, int scan, int oneshot,
-		   int dosyslog, char *config, int test, char *pidfile);
+		   int dosyslog, int test, char *pidfile);
 
-extern int Kill(char *dev, int force);
+extern int Kill(char *dev, int force, int quiet);
 
 extern int CreateBitmap(char *filename, int force, char uuid[16],
 			unsigned long chunksize, unsigned long daemon_sleep,
@@ -277,6 +424,7 @@ extern int CreateBitmap(char *filename, int force, char uuid[16],
 			unsigned long long array_size,
 			int major);
 extern int ExamineBitmap(char *filename, int brief, struct supertype *st);
+extern int bitmap_update_uuid(int fd, int *uuid);
 
 extern int md_get_version(int fd);
 extern int get_linux_version(void);
@@ -289,11 +437,15 @@ extern int get_mdp_major(void);
 extern int dev_open(char *dev, int flags);
 extern int is_standard(char *dev, int *nump);
 
-
-extern mddev_ident_t conf_get_ident(char *conffile, char *dev);
-extern mddev_dev_t conf_get_devs(char *conffile);
-extern char *conf_get_mailaddr(char *conffile);
-extern char *conf_get_program(char *conffile);
+extern int parse_auto(char *str, char *msg, int config);
+extern mddev_ident_t conf_get_ident(char *dev);
+extern mddev_dev_t conf_get_devs(void);
+extern struct createinfo *conf_get_create_info(void);
+extern void set_conffile(char *file);
+extern char *conf_get_mailaddr(void);
+extern char *conf_get_mailfrom(void);
+extern char *conf_get_program(void);
+extern char *conf_get_homehost(void);
 extern char *conf_line(FILE *file);
 extern char *conf_word(FILE *file, int allow_key);
 extern void free_line(char *line);
@@ -353,3 +505,8 @@ extern int open_mddev(char *dev, int autof);
 #define makedev(M,m) (((M)<<8) | (m))
 #endif
 
+/* for raid5 */
+#define ALGORITHM_LEFT_ASYMMETRIC	0
+#define ALGORITHM_RIGHT_ASYMMETRIC	1
+#define ALGORITHM_LEFT_SYMMETRIC	2
+#define ALGORITHM_RIGHT_SYMMETRIC	3
